/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
   SELOR - Servidor de Emails Livre Orientado a Regras                           
   Programado por Lucas Priori                                                   
   Versao: 1.1                                                                   
                                                                                 
   Linux - make linux                                                            
   FreeBSD - gmake freebsd                                                       
                                                                                 
  *******************************************************************            
  Selor e um agente MTA para troca de emails via protocolo SMTP                 
  Copyright (C) 2012,2013,2014  Lucas Priori                                     
                                                                                 
  This program is free software; you can redistribute it and/or                  
  modify it under the terms of the GNU General Public License                    
  as published by the Free Software Foundation; either version 2                 
  of the License, or (at your option) any later version.                         
                                                                                 
  This program is distributed in the hope that it will be useful,                
  but WITHOUT ANY WARRANTY; without even the implied warranty of                 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  
  GNU General Public License for more details.                                   
                                                                                 
  You should have received a copy of the GNU General Public License              
  along with this program; if not, write to the Free Software                    
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<sys/socket.h>
#include<arpa/inet.h>
#include<netinet/in.h>
#include<resolv.h>
#include<netdb.h>

// --
#include "inc/global.h"
#include "inc/selor_utils.h"
#include "inc/selor_dns.h"

#define SPF_None 0
#define SPF_Pass 1
#define SPF_Neutral 2
#define SPF_SoftFail 3
#define SPF_Fail 4
#define SPF_TempError 5
#define SPF_PermError 6


unsigned char buffer_dns[PACKETSZ];
char nome_dns[NS_MAXDNAME];

//======================================================
// Coloca no ponteiro *nome_rev o nome do servidor      
// remoto                                               
// Retornos:                                            
// 0 = Sucesso. End reverso em nome_rev                 
// 1 = Impossivel achar o nome                          
char reverso(unsigned int ip, char *nome_rev, int tam_nome_rev){
 struct hostent* info_rev;
 
 if(!(info_rev=gethostbyaddr((char *)&ip, 4, AF_INET)))return 1;
 selor_strncpy(nome_rev,info_rev->h_name,tam_nome_rev);
 return 0;
}

//======================================================
// Fazer consultas de DNS com base no campo pedido      
// Retornos:                                            
// -1 = Erro de DNS                                     
// 0 = Sucesso                                          
// 1 = Dominio inexistente                              
// 2 = Campo inexistente                                
// A struct dns_resp  preenchida com os dados          
int dns_query(char *dominio, int tipo, unsigned int *num_regs){
 HEADER *cabec;
 unsigned char *ptr_aux1;
 unsigned char *fim_buf;
 int  tam_dn;
 int  tam_buf;
 unsigned int cont1=0;
 unsigned int cont_reconsulta=0;
 unsigned short type;
 unsigned short nq;       // Numero de entradas de questoes  
 unsigned short nr;       // Numero de entradas de respostas 
 unsigned long size;
 
 *num_regs=0;
 if(res_init() < 0)return -1;
 Refazer_consulta:
 tam_buf=res_query(dominio, C_IN, tipo, buffer_dns, PACKETSZ);
 if(tam_buf < 0)return 1;
 fim_buf=buffer_dns+tam_buf;
 cabec=(HEADER *)buffer_dns;
 ptr_aux1=buffer_dns+sizeof(HEADER);
 nq=ntohs(cabec->qdcount);
 nr=ntohs(cabec->ancount);
 for(cont1=0; cont1<nq; cont1++){
  tam_dn=dn_expand(buffer_dns, fim_buf, ptr_aux1, nome_dns, NS_MAXDNAME);
  if(tam_dn < 0)return 1;
  ptr_aux1+=tam_dn+NS_INT16SZ+NS_INT16SZ;
 }
 memset(&dns_resp, 0, sizeof(dns_resp));
 for(cont1=0; cont1<nr && cont1<MAX_query_mx; cont1++){
  tam_dn=dn_expand(buffer_dns, fim_buf, ptr_aux1, nome_dns, NS_MAXDNAME);
  ptr_aux1+=tam_dn;
  GETSHORT(type, ptr_aux1);
  if(type == T_MX){
   ptr_aux1+=8;         //GETSHORT(classe, ptr_aux1);  GETLONG(ttl, ptr_aux1);  GETSHORT(rdlen, ptr_aux1);
   GETSHORT(dns_resp[cont1].pref, ptr_aux1);
   tam_dn=dn_expand(buffer_dns, fim_buf, ptr_aux1, dns_resp[cont1].nome, NS_MAXDNAME);
   ptr_aux1+=tam_dn;
  }else if(type == T_PTR || type == T_CNAME){
   ptr_aux1+=8;         //GETSHORT(classe, ptr_aux1);  GETLONG(ttl, ptr_aux1);  GETSHORT(rdlen, ptr_aux1);
   tam_dn=dn_expand(buffer_dns, fim_buf, ptr_aux1, dns_resp[cont1].nome, NS_MAXDNAME);
   ptr_aux1+=tam_dn;
   if(type == T_CNAME && tipo == T_PTR){
    // Foi pedido um PTR e foi encontrado um CNAME.    
    // Vamos fazer uma nova consulta usando esse CNAME 
    cont_reconsulta++;
    if(cont_reconsulta > 1)return -1;
    dominio=dns_resp[cont1].nome;
    goto Refazer_consulta;
   }
  }else if(type == T_TXT){
   ptr_aux1+=INT16SZ+4;    // 4 = Pular o GETLONG(ttl, ptr_aux1)
   //GETLONG(ttl, ptr_aux1);
   GETSHORT(size, ptr_aux1);
   size=*(ptr_aux1++) & 0xFF;
   strncpy(dns_resp[cont1].nome, (char *)ptr_aux1, size);
   *num_regs=1;
   return 0;
  }
 }
 *num_regs=cont1;
 return 0;
}


//======================================================
// Resolve o endereco IP de determinado host            
// Retorna ZERO caso nao tenha encontrado ou o Numero   
// do IP (unsinged int), e no ponteiro *ip, o IP em     
// formato texto                                        
unsigned int resolver_dom(char *dominio, char *ip, int indice){
 unsigned int ipn=0;
 
 if(!indice && !(host_ips=gethostbyname(dominio)))return 0;
 if(host_ips->h_addr_list[indice] && host_ips->h_addr_list[indice][0]){
  if(ip)strcpy(ip, inet_ntoa(*((struct in_addr *)host_ips->h_addr_list[indice])));
  ipn=((host_ips->h_addr_list[indice][0] & 0xFF)) | ((host_ips->h_addr_list[indice][1] & 0xFF)<<8) | ((host_ips->h_addr_list[indice][2] & 0xFF)<<16) | ((host_ips->h_addr_list[indice][3] & 0xFF)<<24);
  return ipn;
 }else {
  if(ip)ip[0]=0;
  return 0;
 }
}

//======================================================
// Cria uma lista de MX do dominio indexada por pref.   
// Retornos:                                            
// -1 = Erro de DNS                                     
// 0 = Sucesso                                          
// 1 = Dominio inexistente                              
// 2 = Campo inexistente                                
int achar_mx(char *dominio, unsigned int *num_regs){
 unsigned short menor_pref_mx=0;
 int resp;
 unsigned int cont1;
 unsigned int cont2;
 
 dns_num_mxzes=0;     // Zerando a quantidade de MXzes encontrados              
 dns_num_mx=0;        // Zerando o indice do MX para ser usado por prox_mx      
 dns_num_ip=0;        // Zerando o indice do IP do MX para ser usado por prox_mx
 if((resp=dns_query(dominio, T_MX, num_regs)))return resp;
 // -- Indexando MXzes --
 for(cont1=0; cont1<*num_regs; cont1++){
  dns_resp_indexado[cont1].pref=0xFFFF;
  for(cont2=0; cont2<*num_regs; cont2++){
   if(dns_resp[cont2].pref < dns_resp_indexado[cont1].pref){
    dns_resp_indexado[cont1].pref=dns_resp[cont2].pref;
    menor_pref_mx=cont2;
   }
  }
  strcpy(dns_resp_indexado[cont1].nome, dns_resp[menor_pref_mx].nome);
  dns_resp[menor_pref_mx].pref=0xFFFF;
 }
 dns_num_mxzes=*num_regs;
 return resp;
}

//======================================================
// Retorna o proximo IP da lista de MXzes de um dominio 
// Retorna 0 (zero) caso no exista mais numeros de ips 
// para o dominio ou um {unsigned int) como endereo  IP
// e em (char *ip) retorna o IP em forma de texto       
unsigned int prox_mx(char *ip, char ip_ver_alias){
 unsigned int ret_ip;
 
 if(!dns_resp_indexado[dns_num_mx].nome[0])return 0;
 do{
  ret_ip=resolver_dom(dns_resp_indexado[dns_num_mx].nome, ip, dns_num_ip);
  if(!ret_ip){
   // IP Nao encontrado. Vamos ver se existe outro MX
   if(dns_num_mx == dns_num_mxzes)return 0;    // Nao existem mais MXzes
   dns_num_ip=0;
   dns_num_mx++;
  }else {
   if(!ip_ver_alias)dns_num_mx++;        // Para nao procurar pelos IP alias do MX ( Maior velocidade )
   else dns_num_ip++;
  }
 }while(!ret_ip);
 return ret_ip;
}

//======================================================
// Funcao para testar se o IP  considerado local ou nao
// Retornos:                                            
// 0 = Nao e' local                                     
// 1 - IP Local                                         
char ip_local(unsigned int ip){
 unsigned short cont1=0;
 
 if((ip & 0xFF) == 127)return 1;    // 127.x.x.x
 for(cont1=0; cont1<MAX_ips_locais && ips_locais[cont1]; cont1++)if(ip == ips_locais[cont1])return 1;
 return 0;
}

//======================================================
// Funcao para checagem de RBL                          
// Retornos:                                            
// 0 = Nao esta cadastrado em blacklist                 
// Numero RBL onde o IP foi encontrado                  
unsigned short rbl_check(unsigned int ipn, char *rbl_list, char *txt, unsigned int tam_txt){
 char buffer_local[150];
 char *ptr_aux1;
 char *ptr_aux2;
 char *ptr_aux3;
 char *ptr_ip;
 unsigned int rbls=0;
 unsigned int ret_regs;
 
 ptr_ip=(char *)&ipn;
 *txt=0;
 do{
  if((ptr_aux1=strchr(rbl_list, ' ')))*ptr_aux1=0;
  snprintf(buffer_local, sizeof(buffer_local), "%d.%d.%d.%d.%s", (unsigned char)ptr_ip[3], (unsigned char)ptr_ip[2], (unsigned char)ptr_ip[1], (unsigned char)ptr_ip[0], rbl_list);
  if(!dns_query(buffer_local, T_A, &ret_regs)){
   rbls++;
   if(!(*txt)){
    // Buscando TXT com informacoes da RBL
    if(!dns_query(buffer_local, T_TXT, &ret_regs)){
     ptr_aux2=dns_resp[0].nome;
     ptr_aux3=txt;
     do{
      if(*ptr_aux2 == 10 || *ptr_aux2 == 13 || *ptr_aux2 == '"' || *ptr_aux2 == '\'')continue;
      *(ptr_aux3++)=*ptr_aux2;
     }while(*ptr_aux2++);
     *ptr_aux3=0;
    }
   }
  }
  if(ptr_aux1){
   ptr_aux1++;
   ptr_aux1=pular_carac(ptr_aux1, ' ');
   if(*ptr_aux1)rbl_list=ptr_aux1;
   else ptr_aux1=NULL;
  }
 }while(ptr_aux1);
 return rbls;
}

//======================================================
// Funcao para fazer os testes de SPF                   
// Retornos:                                            
// 0 = None                                             
// 1 = Pass                                             
// 2 = Nuetral                                          
// 3 = SoftFail                                         
// 4 = Fail                                             
// 5 = TempError                                        
// 6 = PermError                                        
char spf_check(char *dominio, unsigned int ipn){
 static char spf_recursivo=0;
 char buffer_local[150];
 char spf_string[NS_MAXDNAME];
 char *ptr_aux1;
 char *ptr_aux2;
 char *ptr_aux3;
 char *ptr_aux4;
 char *ptr_ip;
 char ret_spf=0;
 char fim_achado=0;
 char acao=0;
 unsigned short mascara;
 unsigned int spf_ipn;
 unsigned int ret_regs;
 unsigned int cont1;
 unsigned int cont2;
 
 
 if(!dominio)return 0;
 if(dns_query(dominio, T_TXT, &ret_regs))return SPF_None;
 strcpy(spf_string, dns_resp[0].nome);
 ptr_ip=(char *)&ipn;
 if(cmp_str(spf_string, "v=spf1 ", 7))return SPF_None;
 ptr_aux1=spf_string+7;
 if(strchr(dns_resp[0].nome, '%')){
  // Macros nao sao suportadas por enquanto
  return SPF_PermError;
 }
 do{
  ptr_aux2=ptr_aux1=pular_carac(ptr_aux1, ' ');
  while(*ptr_aux2 != ' ' && *ptr_aux2)ptr_aux2++;
  if(!*ptr_aux2)fim_achado=1;
  else *ptr_aux2=0;
  // -- Pegando a ACAO --
  if(*ptr_aux1 == '-')acao=SPF_Fail;
  else if(*ptr_aux1 == '~')acao=SPF_SoftFail;
  else if(*ptr_aux1 == '?')acao=SPF_Neutral;
  else {
   if(*ptr_aux1 != '+')ptr_aux1--;
   acao=SPF_Pass;
  }
  ptr_aux1++;
  // -- Mecanismo --
  mascara=32;
  if((ptr_aux3=strchr(ptr_aux1, '/'))){
   *(ptr_aux3++)=0;
   mascara=atoi(ptr_aux3);
  }
  cont1=0;
  // -- IP4 --
  if(*ptr_aux1 == 'i' && ptr_aux1[1] == 'p' && ptr_aux1[2] == '4' && ptr_aux1[3] == ':'){
   ptr_aux1+=4;
   if(!*ptr_aux1)return SPF_PermError;
   if((spf_ipn=inet_addr(ptr_aux1)) == 0xFFFFFFFF)return SPF_PermError;
   if(tmasc(ipn, spf_ipn, mascara))return acao;
   
  // -- MX --
  }else if(*ptr_aux1 == 'm' && ptr_aux1[1] == 'x'){
   ptr_aux1+=2;
   if(*ptr_aux1 == ':')ptr_aux3=++ptr_aux1;
   else ptr_aux3=dominio;
   if(!dns_query(ptr_aux3, T_MX, &ret_regs)){
    for(cont1=0; cont1<ret_regs; cont1++){
     cont2=0;
     while((spf_ipn=resolver_dom(dns_resp[cont1].nome, NULL, cont2))){
      if(tmasc(ipn, spf_ipn, mascara))return acao;
      cont2++;
     }
    }
   }
   
  // -- A --
  }else if(*ptr_aux1 == 'a' && (ptr_aux1[1] == ':' || !ptr_aux1[1])){
   if(*(++ptr_aux1) == ':')ptr_aux3=++ptr_aux1;
   else ptr_aux3=dominio;
   while((spf_ipn=resolver_dom(ptr_aux3, NULL, cont1))){
    if(tmasc(ipn, spf_ipn, mascara))return acao;
    cont1++;
   }
   
  // -- PTR --
  }else if(*ptr_aux1 == 'p' && ptr_aux1[1] == 't' && ptr_aux1[2] == 'r'){
   ptr_aux1+=3;
   if(*ptr_aux1 == ':')ptr_aux3=++ptr_aux1;
   else ptr_aux3=dominio;
   snprintf(buffer_local, sizeof(buffer_local), "%d.%d.%d.%d.in-addr.arpa", (unsigned char)ptr_ip[3], (unsigned char)ptr_ip[2], (unsigned char)ptr_ip[1], (unsigned char)ptr_ip[0]);
   if(!dns_query(buffer_local, T_PTR, &ret_regs)){
    for(cont1=0; cont1<ret_regs; cont1++){
     cont2=0;
     while((spf_ipn=resolver_dom(dns_resp[cont1].nome, NULL, cont2++))){
      if(ipn == spf_ipn){
       if(!strcmp(ptr_aux3, dns_resp[cont1].nome))return acao;
       if((ptr_aux4=proc_str(dns_resp[cont1].nome, ptr_aux3)) && *(--ptr_aux4) == '.')return acao;
      }
     }
    }
   }
   
  // -- INCLUDE --
  }else if(!strncmp(ptr_aux1, "include:", 8)){
   if(spf_recursivo > MAX_spf_recursividade)return SPF_PermError;       // Para impedir loops de SPF
   spf_recursivo++;
   ptr_aux1+=8;
   ret_spf=spf_check(ptr_aux1, ipn);
   spf_recursivo--;
   if(ret_spf == 1)return acao;
   if(!ret_spf)return SPF_PermError;
   if(ret_spf > 4)return ret_spf;

  // -- REDIRECT --
  }else if(!strncmp(ptr_aux1, "redirect=", 9)){
   if(spf_recursivo > MAX_spf_recursividade)return SPF_PermError;       // Para impedir loops de SPF
   spf_recursivo++;
   ptr_aux1+=9;
   ret_spf=spf_check(ptr_aux1, ipn);
   spf_recursivo--;
   if(ret_spf == SPF_None)return SPF_PermError;
   return ret_spf;
   
  // -- ALL --
  }else if(*ptr_aux1 == 'a' && ptr_aux1[1] == 'l' && ptr_aux1[2] == 'l')return acao;
  else {
   // -- Mecanismo nao reconhecido --
   return SPF_PermError;
  }
  ptr_aux1=++ptr_aux2;
 }while(!fim_achado);
 return SPF_Neutral;
}

















